-------------------------------------------------------------------------------
--! @file   AEZ_E.vhd
--! @brief  AEZ_E module
--!
--!         param_j
--!             determines the precalculation operation
--!             and the subsequent operation of the start signal

--!             -----|---------|-
--!             mode | (j, i)  |
--!                  |         |
--!             -----|---------|-
--!               0  | ( 0, x) |
--!               1  | ( 1, x) |
--!               2  | ( 2, x) |
--!               3  | ( 3, 1) |
--!               4  | ( 4, 0) |
--!               5  | ( 5, x) |
--!               7  | (-1, x) |
--!
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.AES_pkg.all;

entity AEZ_E is
    port    (
        clk                 : in  std_logic;
        rst                 : in  std_logic;
        --! Input
        key                 : in  std_logic_vector( 32          -1 downto 0);
        xx                  : in  std_logic_vector(128          -1 downto 0);
        --! Pre-Processor Control
        key_valid           : in  std_logic;
        key_ready           : out std_logic;
        --! Control
        ctrl                : in  std_logic_vector( 14          -1 downto 0);
        --! Status
        init_key_done       : out std_logic;
        almost_done         : out std_logic;
        done                : out std_logic;
        ready               : out std_logic;
        --! Output
        yy                  : out std_logic_vector(128          -1 downto 0)
    );
end entity AEZ_E;

architecture structure of AEZ_E is
    constant E_RND                  : integer := 4;
    constant E_RND_BIG              : integer := 10;

    --! Control signals
    --!     Input
    signal clr_state                : std_logic;
    signal start                    : std_logic;
    signal init_key                 : std_logic;
    signal param_j                  : std_logic_vector(  3          -1 downto 0);
    signal param_i_lo               : std_logic_vector(  3          -1 downto 0);
    signal param_i_hi               : std_logic_vector(  5          -1 downto 0);
    --!     Status
    signal compute_round_r          : std_logic_vector(  5          -1 downto 0);
    signal round                    : std_logic_vector(  5          -1 downto 0);
    --!     from controller
    signal almost_done_s            : std_logic;
    signal done_s                   : std_logic;
    signal init_key_done_s          : std_logic;
    signal en_key                   : std_logic;
    signal en_key_init              : std_logic;
    signal param_i_lo_r             : std_logic_vector(  3          -1 downto 0);
    signal param_j_r                : std_logic_vector(  3          -1 downto 0);
    signal dcalc_en_r               : std_logic_vector(  3          -1 downto 0);
    signal en_state                 : std_logic;
    signal en_ram                   : std_logic;
    signal en_round                 : std_logic;
    signal clr_round                : std_logic;
    signal sel_key_init             : std_logic_vector(  2          -1 downto 0);
    signal sel_init_lj              : std_logic_vector(  3          -1 downto 0);
    signal sel_round_type           : std_logic;
    signal done_ram_init            : std_logic;
    --!     derived
    signal is_big_e_pre             : std_logic;
    signal is_big_e                 : std_logic;
    signal is_isetup                : std_logic;
    signal iram_sel                 : std_logic_vector(  5          -1 downto 0);
    signal iram_sel_r               : std_logic_vector(  5          -1 downto 0);
    signal sel_roundkey             : std_logic_vector(  2          -1 downto 0);
    signal rom_roundkey_addr        : std_logic_vector(  5          -1 downto 0);
    signal compute_addr             : std_logic_vector(  6          -1 downto 0);
    signal ram_addr                 : std_logic_vector(  6          -1 downto 0);
    --!     State
    type    state_type is (S_IDLE, S_LD_KEY, S_INIT, S_SETUP, S_SETUP_I, S_COMPUTE);
    signal state                    : state_type;
    signal nstate                   : state_type;

    --! Data signals
    --!     signals
    signal key_r                    : std_logic_vector(384          -1 downto 0);
    signal key_ii                   : std_logic_vector(128          -1 downto 0);
    signal key_jj                   : std_logic_vector(128          -1 downto 0);
    signal key_ll                   : std_logic_vector(128          -1 downto 0);
    signal key_init_r               : std_logic_vector(128          -1 downto 0);
    signal selected_key             : std_logic_vector(128          -1 downto 0);
    signal double_key               : std_logic_vector(128          -1 downto 0);
    signal oo_aes                   : std_logic_vector(128          -1 downto 0);
    signal ram_data                 : std_logic_vector(128          -1 downto 0);
    signal roundkey                 : std_logic_vector(128          -1 downto 0);
    signal init_xor                 : std_logic_vector(128          -1 downto 0);
    --!     registers
    signal delta_enc_r              : std_logic_vector(128          -1 downto 0);
    signal state_r                  : std_logic_vector(128          -1 downto 0);
    --!     mapped signals
    signal state_r_mapped           : t_AES_state;
    signal roundkey_mapped          : t_AES_state;
    signal oo_aes_mapped            : t_AES_state;


    --! RAM
    --! Ram[0-7 ]  = i*L = [0-7]*L
    --!     Note: i will never exceed 7
    --! Ram[8-15]  = i*J = [0-7]*J
    --!     Note: only j=0-5 is used for this specific implementation
    --! Ram[16-32] = (2^up(i/8))*I = [2^(0-15)]*I
    type ram_type is array(0 to 63) of std_logic_vector(128         -1 downto 0);
    signal ram : ram_type;

    --! Rom
    type rom_type is array(0 to 31) of std_logic_vector(  2         -1 downto 0);
    constant rom_roundkey : rom_type :=
        ("01", "00", "10", "11", "00", "00", "00", "00", --![ 0- 3] = AEZ4
         "00", "00", "00", "00", "00", "00", "00", "00", --![ 4-15] = XX (00)
         "00", "01", "10", "00", "01", "10", "00", "01", --![32-42] = AEZ10
         "10", "00", "00", "00", "00", "00", "00", "00"); --![43-63] = XX (00)

    --! Function
    function double_block(xx: std_logic_vector(128-1 downto 0))
    return std_logic_vector is
        variable shifted_xx : std_logic_vector(128-1 downto 0);
        variable msb        : std_logic;
        variable output     : std_logic_vector(128-1 downto 0);
    begin
        shifted_xx := xx(126 downto 0) & '0';
        msb        := xx(127);
        output :=   (shifted_xx(127 downto 8)) &
                    (shifted_xx(7) xor msb)    &
                    (shifted_xx(6 downto 3))   &
                    (shifted_xx(2) xor msb)    &
                    (shifted_xx(1) xor msb)    &
                    (shifted_xx(0) xor msb);
        return output;
    end function double_block;
begin
    start      <= ctrl( 0);
    init_key   <= ctrl( 1);
    clr_state  <= ctrl( 2);
    param_j    <= ctrl( 5 downto 3);
    param_i_lo <= ctrl( 8 downto 6);
    param_i_hi <= ctrl(13 downto 9);

    --! ==============
    --! Datapath
    --! ==============
    --! Internal signals
    key_ii              <= key_r(384-1 downto 256);
    key_jj              <= key_r(256-1 downto 128);
    key_ll              <= key_r(128-1 downto   0);

    --! Registers
    pReg: process(clk)
    begin
        if rising_edge(clk) then
            --! Key register
            if (en_key = '1') then
                key_r <= key_r(351 downto 0) & key;
            end if;
            --! Key doubling initialization register
            if (en_key_init = '1') then
                key_init_r <= selected_key;
            end if;
            --! Delta pre-computation register
            if (start = '1') then
                delta_enc_r <= (others => '0');
            elsif (dcalc_en_r(0) = '1') then
                delta_enc_r <= delta_enc_r xor ram_data;
            end if;
            --! State register
            if (clr_state = '1') then
                state_r <= (others => '0');
            elsif (start = '1') then
                state_r <= delta_enc_r xor xx;
            elsif (en_state = '1') then
                state_r <= oo_aes;
            end if;
            --! Key doubling RAM
            if (en_ram = '1') then
                ram(to_integer(unsigned(ram_addr))) <= key_init_r xor init_xor;
            end if;
        end if;
    end process;

    --! ================ Previous block operation
    --! Key doubling pre-computation
    with sel_key_init select
    selected_key <= key_ii   when "01",
                    key_jj   when "10",
                    key_ll   when "11",
                    double_key when others;
    double_key   <= double_block(key_init_r);

    with sel_init_lj(2 downto 0) select
    init_xor <= key_ll                          when "001",
                double_block(key_ll)            when "010",
                double_block(key_ll) xor key_ll when "011",
                key_jj                          when "101",
                double_block(key_jj)            when "110",
                double_block(key_jj) xor key_jj when "111",
                (others => '0')                 when others;

    ram_data <= ram(to_integer(unsigned(ram_addr)));

    --! ================ Current block operation
    --! I and J RAM addr selection
    with round(1 downto 0) select
    compute_addr <= ("000" & param_i_lo_r) when "00",
                    ("001" & param_j_r) when "01",
                    ('1' & iram_sel_r    ) when others;
    is_isetup <= '1' when (state = S_SETUP_I) else '0';
    ram_addr <= is_isetup & round when (state = S_SETUP or state = S_SETUP_I) else compute_addr;

    --! Main
    uAes: entity work.AES_Round(basic)
    port map (din       => state_r_mapped,
              rkey      => roundkey_mapped,
              dout_fdb  => oo_aes_mapped,
              dout      => open);

    u_map_din: entity work.AES_map(structure)
    port map (ii => state_r,
              oo => state_r_mapped);
    u_map_key: entity work.AES_map(structure)
    port map (ii => roundkey,
              oo => roundkey_mapped);
    u_invmap: entity work.AES_invmap(structure)
    port map (ii => oo_aes_mapped,
              oo => oo_aes);

    with sel_roundkey(1 downto 0) select
    roundkey <= key_ii          when "00",
                key_jj          when "01",
                key_ll          when "10",
                (others => '0') when others;

    rom_roundkey_addr <= sel_round_type & round(3 downto 0);
    sel_roundkey      <= rom_roundkey(to_integer(unsigned(rom_roundkey_addr)));

    --! Output
    yy     <= state_r;

    --! ==============
    --! Controller
    --! ==============
    --! A block that calculates address for i-ram
    gIramSel: block
        signal need_roundup : std_logic;
        signal rup_block : unsigned(4 downto 0);
    begin
        need_roundup <= (param_i_lo(0) or param_i_lo(1) or param_i_lo(2));
        rup_block <= unsigned(param_i_hi) + (need_roundup & "");
        iram_sel <= std_logic_vector(rup_block);
    end block;

    pCtrlReg: process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then
                state <= S_IDLE;
            else
                state <= nstate;
            end if;

            if (clr_round = '1') then
                round <= (others => '0');
            elsif (en_round = '1') then
                round <= std_logic_vector(unsigned(round) + natural(1));
            end if;

            --! Operation related status
            if (rst = '1') then
                is_big_e <= '1';
            elsif (start = '1') then
                --! Prepare
                param_i_lo_r <= param_i_lo;
                param_j_r    <= param_j;
                iram_sel_r   <= iram_sel;
                if (param_j = "111") then
                    is_big_e_pre <= '1';
                else
                    is_big_e_pre <= '0';
                end if;
                --! Compute
                is_big_e <= is_big_e_pre;
                if (is_big_e_pre = '1') then
                    compute_round_r <= std_logic_vector(to_unsigned(E_RND_BIG-1,
                                            compute_round_r'length));
                else
                    compute_round_r <= std_logic_vector(to_unsigned(E_RND-1,
                                            compute_round_r'length));
                end if;
            end if;

            if (start = '1') then
                if (param_j = "111") then
                    dcalc_en_r <= "001";
                else
                    dcalc_en_r <= "111";
                end if;
            elsif (en_round = '1') then
                dcalc_en_r <= '0' & dcalc_en_r(2 downto 1);
            end if;

            --! Output status
            if (rst = '1') then
                ready <= '1';
            elsif (start = '1') then
                ready <= '0';
            elsif (done_s = '1') then
                ready <= '1';
            end if;
            done            <= done_s;
            almost_done     <= almost_done_s;
            init_key_done   <= init_key_done_s;
        end if;
    end process;

    sel_round_type <= is_big_e;

    pCtrlComb: process(state, init_key, round, start, compute_round_r)
    begin
        --! External
        key_ready       <= '0';
        init_key_done_s <= '0';
        done_s          <= '0';
        --! internal
        clr_round       <= '0';
        en_key          <= '0';
        en_round        <= '0';
        en_ram          <= '0';
        en_key_init     <= '0';
        en_state        <= '0';
        sel_key_init    <= "00";
        done_ram_init   <= '0';
        almost_done_s   <= '0';

        sel_init_lj     <= "000";

        case state is
            when S_IDLE =>
                clr_round <= '1';

                if (init_key = '1') then
                    nstate <= S_LD_KEY;
                elsif (start = '1') then
                    nstate <= S_COMPUTE;
                    en_state       <= '1';
                end if;

            when S_LD_KEY =>
                key_ready <= '1';
                if key_valid = '1' then
                    en_key       <= '1';
                    en_round     <= '1';
                    if (round = "01011") then
                        clr_round <= '1';
                        nstate    <= S_INIT;
                    end if;
                end if;

            when S_INIT =>
                nstate       <= S_SETUP;
                sel_key_init <= "11";
                en_key_init  <= '1';

            when S_SETUP =>
                --! Populate L and J RAM
                en_round     <= '1';
                en_ram       <= '1';
                if (round(2 downto 0) = "001" or round(2 downto 0) = "011" or round(2 downto 0) = "111") then
                    en_key_init <= '1';
                end if;
                if (round = "00111") then
                    sel_key_init <= "10";
                elsif (round = "01111") then
                    sel_key_init <= "01";
                    clr_round <= '1';
                    nstate    <=  S_SETUP_I;
                end if;
                if (round(2 downto 0) = "000" or round(2 downto 0) = "011" or round(2 downto 0) = "101") then
                    sel_init_lj(1 downto 0)  <= "01";
                elsif (round(2 downto 0) = "110") then
                    sel_init_lj(1 downto 0)  <= "10";
                elsif (round(2 downto 0) = "111") then
                    sel_init_lj(1 downto 0)  <= "11";
                end if;
                sel_init_lj(2) <= round(3);

            when S_SETUP_I =>
                --! Populate I RAM
                en_key_init  <= '1';
                en_round     <= '1';
                en_ram       <= '1';
                if (round = "11111") then
                    done_ram_init   <= '1';
                    nstate          <= S_IDLE;
                    done_s          <= '1';
                    init_key_done_s <= '1';
                end if;
                if (round = "11110") then
                    almost_done_s <= '1';
                end if;

            when S_COMPUTE =>
                en_state <= '1';
                en_round <= '1';
                if (round = compute_round_r) then
                    nstate    <= S_IDLE;
                    clr_round <= '1';
                    done_s    <= '1';
                end if;
                if (unsigned(round) = unsigned(compute_round_r)-natural(1)) then
                    almost_done_s <= '1';
                end if;
        end case;
    end process;
end architecture structure;
