-------------------------------------------------------------------------------
--! @file       CipherCore_TP.vhd
--! @brief      CipherCore_TP for AEZ
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;
use work.AEAD_pkg.all;

entity CipherCore_TP is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_FW            : integer := 128;   --! Two-Pass FIFO
        G_DBLK_SIZE     : integer := 128;   --! Data
        G_KEY_SIZE      : integer := 32;    --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 4;
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := TWO_PASS_MAX
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic;
        --! Second-pass FIFO
        fdi_data        : in  std_logic_vector(G_FW             -1 downto 0);
        fdi_valid       : in  std_logic;
        fdi_ready       : out std_logic;
        fdo_data        : out std_logic_vector(G_FW             -1 downto 0);
        fdo_valid       : out std_logic;
        fdo_ready       : in  std_logic
    );
end entity CipherCore_TP;

architecture structure of CipherCore_TP is
    signal ctrl         : std_logic_vector(38           -1 downto 0);
    signal ctrl_e       : std_logic_vector(14           -1 downto 0);
    signal rtr          : std_logic_vector( 4           -1 downto 0);
    signal data_size    : std_logic_vector( 4           -1 downto 0);
    signal edone        : std_logic;
    signal eready       : std_logic;
    signal init_key_done: std_logic;
begin
    u_cc_dp:
    entity work.CipherCore_TP_Datapath(structure)
    port map (
        clk                 => clk              ,

        --! Input Processor
        bdi                 => bdi              ,
        key                 => key              ,
        key_valid           => key_valid        ,
        key_ready           => key_ready        ,

        --! Controller
        init_key_done       => init_key_done    ,
        ctrl                => ctrl             ,
        ctrl_e              => ctrl_e           ,
        rtr                 => rtr              ,
        data_size           => data_size        ,
        edone               => edone            ,
        eready              => eready           ,

        --! 2nd-pass FIFO
        fdi_data            => fdi_data         ,
        fdo_data            => fdo_data         ,

        --! Output Processor
        bdo                 => bdo              ,
        msg_auth_valid      => msg_auth_valid
    );

    u_cc_ctrl:
    entity work.CipherCore_TP_Control(behavior)
    port map (
        clk                 => clk              ,
        rst                 => rst              ,

        --! Input Processor
        key_update          => key_update       ,
        bdi_valid           => bdi_valid        ,
        bdi_ready           => bdi_ready        ,
        decrypt             => decrypt          ,
        len_a               => bdi(111 downto 96),
        len_d               => bdi( 79 downto 64),

        --! Datapath
        init_key_done       => init_key_done    ,
        ctrl                => ctrl             ,
        ctrl_e              => ctrl_e           ,
        rtr                 => rtr              ,
        data_size           => data_size        ,
        edone               => edone            ,
        eready              => eready           ,

        --! Second pass FIFO
        fdo_valid           => fdo_valid        ,
        fdi_ready           => fdi_ready        ,
        fdo_ready           => fdo_ready        ,
        fdi_valid           => fdi_valid        ,

        --! Output Processor
        msg_auth_done       => msg_auth_done    ,
        bdo_valid           => bdo_valid        ,
        bdo_ready           => bdo_ready
    );
end structure;