-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @brief      Datapath unit for AEZ
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;

entity CipherCore_TP_Datapath is
    port    (        
        clk             : in  std_logic;

        --! Pre-Processor
        --!     Data
        bdi             : in  std_logic_vector(128          -1 downto 0);
        key             : in  std_logic_vector( 32          -1 downto 0);        
        --!     Control
        key_valid       : in  std_logic;
        key_ready       : out std_logic;
        
        --! 2nd-pass FIFO
        fdi_data        : in  std_logic_vector(128          -1 downto 0);
        fdo_data        : out std_logic_vector(128          -1 downto 0);
        
        --! Control
        init_key_done   : out std_logic;
        ctrl            : in  std_logic_vector( 38          -1 downto 0);
        ctrl_e          : in  std_logic_vector( 14          -1 downto 0);
        rtr             : in  std_logic_vector(  4          -1 downto 0);
        data_size       : in  std_logic_vector(  4          -1 downto 0);
        edone           : out std_logic;
        eready          : out std_logic;

        --! Post-Processor
        bdo             : out std_logic_vector(128          -1 downto 0);
        msg_auth_valid  : out std_logic
    );
end entity CipherCore_TP_Datapath;

architecture structure of CipherCore_TP_Datapath is
    --! Control signals
    --!     Top
    signal en_out               : std_logic;
    signal en_ll                : std_logic;
    signal en_xy                : std_logic;
    signal en_hash              : std_logic;
    signal en_ss                : std_logic;
    signal en_tmp               : std_logic;
    signal sel_mask             : std_logic;
    signal sel_pad              : std_logic;
    signal sel_tau              : std_logic;    
    signal sel_ll               : std_logic;    
    signal sel_rothb            : std_logic;
    signal sel_ldata            : std_logic_vector(  3          -1 downto 0);
    signal sel_ldata_xor        : std_logic_vector(  2          -1 downto 0);
    signal sel_data             : std_logic_vector(  3          -1 downto 0);
    signal sel_tiny             : std_logic;
    signal sel_xy               : std_logic_vector(  2          -1 downto 0);
    signal sel_tag              : std_logic;
    signal sel_output           : std_logic;
    signal sel_fifo             : std_logic;    
    signal tinyround            : std_logic_vector(  3          -1 downto 0);
    signal mask_halfbyte        : std_logic;    
    signal sel_rot              : std_logic_vector(  2          -1 downto 0);
    signal clr_init             : std_logic;
    signal sel_sout             : std_logic;
    signal sel_tmp_xor          : std_logic;
    signal sel_e                : std_logic;
    signal sel_otag             : std_logic;
    signal en_tag               : std_logic;
    --!     Derived
    signal sel_exor             : std_logic_vector(  2          -1 downto 0);
    signal sel_tmp_xor1         : std_logic_vector(  2          -1 downto 0);

    --! Data signals
    --!     Registers    
    signal exp_tag_r            : std_logic_vector(128          -1 downto 0);
    signal out_r                : std_logic_vector(128          -1 downto 0);
    signal ll_r                 : std_logic_vector(128          -1 downto 0);
    signal xy_r                 : std_logic_vector(128          -1 downto 0);
    signal hash_r               : std_logic_vector(128          -1 downto 0);
    signal ss_r                 : std_logic_vector(128          -1 downto 0);
    signal tmp_r                : std_logic_vector(128          -1 downto 0);
    signal masks                : std_logic_vector(16           -1 downto 0);
    signal padloc               : std_logic_vector(16           -1 downto 0);

    --!     Signals
    signal from_fifo            : std_logic_vector(128          -1 downto 0);
    signal to_fifo              : std_logic_vector(128          -1 downto 0);
    
    signal rotin                : std_logic_vector(128          -1 downto 0);
    signal rotout               : std_logic_vector(128          -1 downto 0);
    signal data                 : std_logic_vector(128          -1 downto 0);
    signal xor_main             : std_logic_vector(128          -1 downto 0);
    signal to_ll                : std_logic_vector(128          -1 downto 0);
    signal to_ll_xor1           : std_logic_vector(128          -1 downto 0);
    signal to_ll_xor2           : std_logic_vector(128          -1 downto 0);
    signal to_tmp               : std_logic_vector(128          -1 downto 0);
    signal to_tmp_xor           : std_logic_vector(128          -1 downto 0);
    signal rothbout             : std_logic_vector(128          -1 downto 0);
    signal masked_data          : std_logic_vector(128          -1 downto 0);
    signal padded_data          : std_logic_vector(128          -1 downto 0);
    signal tiny_sel             : std_logic_vector(128          -1 downto 0);
    signal aez_ei               : std_logic_vector(128          -1 downto 0);
    signal aez_eo               : std_logic_vector(128          -1 downto 0);
    signal to_xy_xor            : std_logic_vector(128          -1 downto 0);
    signal to_xy                : std_logic_vector(128          -1 downto 0);
    signal tag_s                : std_logic_vector(128          -1 downto 0);
    signal sel_out              : std_logic_vector(2            -1 downto 0);
    signal msg_auth_valid_prf   : std_logic;
    signal msg_auth_valid_data  : std_logic;

    type lookup_table is array (0 to 15) of std_logic_vector(16 -1 downto 0);
    constant masks_lookup   : lookup_table := (
        x"0000", x"8000", x"C000", x"E000", x"F000", x"F800", x"FC00", x"FE00",
        x"FF00", x"FF80", x"FFC0", x"FFE0", x"FFF0", x"FFF8", x"FFFC", x"FFFE");
    constant padloc_lookup  : lookup_table := (
        x"8000", x"4000", x"2000", x"1000", x"0800", x"0400", x"0200", x"0100",
        x"0080", x"0040", x"0020", x"0010", x"0008", x"0004", x"0002", x"0001");
begin
    from_fifo      <= fdi_data;
    fdo_data       <= to_fifo ;
        
    en_out         <= ctrl( 0);
    en_ll          <= ctrl( 1);
    en_xy          <= ctrl( 2);
    en_hash        <= ctrl( 3);
    en_ss          <= ctrl( 4);
    en_tmp         <= ctrl( 5);
    sel_mask       <= ctrl( 6);
    sel_pad        <= ctrl( 7);
    sel_tau        <= ctrl( 8);
    sel_ll         <= ctrl( 9);
    sel_ldata      <= ctrl(13 downto 11);
    sel_ldata_xor  <= ctrl(15 downto 14);
    sel_data       <= ctrl(18 downto 16);
    sel_tiny       <= ctrl(19);
    sel_xy         <= ctrl(21 downto 20);
    sel_tag        <= ctrl(22);
    sel_output     <= ctrl(23);
    sel_fifo       <= ctrl(24);
    sel_rothb      <= ctrl(25);
    tinyround      <= ctrl(28 downto 26);
    mask_halfbyte  <= ctrl(29);
    sel_rot        <= ctrl(31 downto 30);
    clr_init       <= ctrl(32);
    sel_sout       <= ctrl(33);
    sel_tmp_xor    <= ctrl(34);
    sel_e          <= ctrl(35);
    sel_otag       <= ctrl(36);
    en_tag         <= ctrl(37);

    --! Registers
    pReg: process(clk)
    begin
        if rising_edge(clk) then            
            if (en_ll = '1') then
                ll_r    <= to_ll;
            end if;
            if (en_xy = '1') then
                xy_r    <= to_xy;
            end if;
            if (en_hash = '1') then
                hash_r  <= xy_r;
            end if;
            if (clr_init = '1') then
                ss_r    <= (others => '0');
            elsif (en_ss = '1') then
                ss_r    <= xor_main xor ss_r;
            end if;
            if (en_tmp = '1') then
                tmp_r   <= to_tmp;
            end if;
            if (en_out = '1') then
                out_r   <= xor_main;
            end if;
            if (en_tag = '1') then
                exp_tag_r <= bdi;
            end if;

            if (sel_mask = '0') then
                masks <= (others => '1');                
            else
                masks <= masks_lookup(to_integer(unsigned(data_size)));
            end if;
            padloc <= padloc_lookup(to_integer(unsigned(data_size)));
        end if;
    end process;

    --! auxiliary
    with sel_rot select
    rotin <= bdi                when "01",
             tmp_r              when "10",
             exp_tag_r          when "11",  --! Tag
             (others => '0')    when others;

    uBarrelRot1: entity work.bshift_byte(structure)
        generic map (G_BYTE => 16, G_LOG2_BYTE => 4, G_LEFT => 1, G_ROTATE => 1, G_SHIFT1 => 0)
        port map    (ii  => rotin, rtr => rtr, oo => rotout);
    
    sel_tmp_xor1 <= sel_e & sel_rothb;
    with sel_tmp_xor1 select    
    rothbout <= (rotout(123 downto 0) & rotout(127 downto 124))    when "01",
                aez_eo                                             when "10",
                rotout                                             when others;

    to_tmp_xor <= (others => '0') when sel_tmp_xor = '1' else padded_data;    
                  
    to_tmp <= to_tmp_xor xor rothbout;

    --! LL register
    with sel_ldata select
    to_ll_xor1  <=  bdi             when "000",
                    from_fifo       when "001",
                    tmp_r           when "010",
                    xor_main        when "011",
                    (others => '0') when others;

    with sel_ldata_xor select
    to_ll_xor2  <=  xy_r            when "01",
                    aez_eo          when "10",
                    (others => '0') when others;


    to_ll      <= to_ll_xor1 xor to_ll_xor2;

    --! Main data selection
    with sel_data select
    data <= bdi                             when "000",
            from_fifo                       when "001",
            (7 => sel_tau, others => '0')   when "010",
            bdi(127 downto 32) & x"80000000" when "011",
            tmp_r                           when "100",
            ss_r                            when "101",
            xy_r                            when "110",     --! hash (sum)
            xor_main                        when others;    --! Tau or empty

    --! Masking and Padding unit
    gPad: for i in 15 downto 0 generate
        gBit: for j in 7 downto 0 generate
            g7: if j = 7 generate
                masked_data(8*i+j) <=  (data(8*i+j) and masks(i))
                                    or (data(8*i+j) and padloc(i) and     mask_halfbyte);
                padded_data(8*i+j) <=  masked_data(8*i+j)
                                    or ((                padloc(i) and not mask_halfbyte) and sel_pad);
            end generate;
            g654: if (j < 7 and j > 3) generate
                masked_data(8*i+j) <=  (data(8*i+j) and masks(i))
                                    or (data(8*i+j) and padloc(i)   and mask_halfbyte);
                padded_data(8*i+j) <=  masked_data(8*i+j);
            end generate;
            g3: if j = 3 generate
                masked_data(8*i+j) <=  (data(8*i+j) and masks(i));
                padded_data(8*i+j) <=  masked_data(8*i+j)
                                    or ((                padloc(i) and mask_halfbyte) and sel_pad);
            end generate;
            g210: if (j < 3) generate
                masked_data(8*i+j) <=  (data(8*i+j) and masks(i));
                padded_data(8*i+j) <=  masked_data(8*i+j);
            end generate;
        end generate;
    end generate;

    sel_exor <= sel_sout & sel_tiny;
    with sel_exor select
    tiny_sel <= hash_r(127 downto 3) & (hash_r(2 downto 0) xor tinyround)   when "01",
                aez_eo                                                      when "10",
                (others => '0')                                             when others;
    
    aez_ei     <= padded_data xor tiny_sel;

    --! Core
    uEncrypt: entity work.AEZ_E(structure)
        port map (rst           => clr_init, 
                  clk           => clk,  
                  --! Input
                  key           => key, 
                  xx            => aez_ei, 
                  --! Control
                  key_ready     => key_ready,
                  key_valid     => key_valid,
                  ctrl          => ctrl_e,
                  --! Status
                  init_key_done => init_key_done,
                  --!   Note: the control logic are registered 
                  --!         so take penultimate signal
                  almost_done   => edone,   
                  ready         => eready,
                  --! Output
                  yy            => aez_eo);

    xor_main <= aez_eo xor ll_r;

    to_xy_xor <= ll_r when sel_ll = '1' else (others => '0');
    with sel_xy select
    to_xy <= aez_eo xor xy_r xor to_xy_xor   when "00",
             hash_r             when "01",
             tag_s              when "10",
             (others => '0')    when others;

    --! Tag
    tag_s   <= aez_eo   when sel_tag = '1'   else tmp_r;
    msg_auth_valid_prf  <= '1' when (aez_eo = exp_tag_r) else '0';
    msg_auth_valid_data <= '1' when (unsigned(tag_s) = natural(0)) else '0';
    msg_auth_valid      <= msg_auth_valid_prf when (sel_tag = '1') else msg_auth_valid_data;

    --! output
    sel_out <= sel_otag & sel_output;
    with sel_out select
    bdo <= xor_main when "00",
             out_r when   "01",
             tag_s when   "10",
             xy_r  when   others;

    --! Fifo output
    to_fifo <= xor_main;

end architecture structure;